@@ -9,9 +9,12 @@ from collections.abc import Awaitable, Mapping
 from functools import cached_property
 from typing import TYPE_CHECKING, Any, Literal, Optional
 
+import msgspec
+
 from marimo import _loggers
-from marimo._ast.sql_visitor import SQLVisitor
+from marimo._ast.sql_visitor import SQLRef, SQLVisitor
 from marimo._ast.visitor import ImportData, Language, Name, VariableData
+from marimo._messaging.msgspec_encoder import asdict
 from marimo._runtime.exceptions import MarimoRuntimeException
 from marimo._types.ids import CellId_t
 from marimo._utils.deep_merge import deep_merge
@@ -27,8 +30,8 @@ if TYPE_CHECKING:
     from marimo._output.hypertext import Html
 
 
-@dataclasses.dataclass
-class CellConfig:
+# TODO: Use rename="camel" for consistency in JSON-encoding
+class CellConfig(msgspec.Struct):
     """
     Internal representation of a cell's configuration.
     This is not part of the public API.
@@ -55,7 +58,7 @@ class CellConfig:
         return config
 
     def asdict(self) -> dict[str, Any]:
-        return dataclasses.asdict(self)
+        return asdict(self)
 
     def asdict_without_defaults(self) -> dict[str, Any]:
         return {
@@ -73,16 +76,14 @@ class CellConfig:
         `update` can be a partial config or a CellConfig
         """
         if isinstance(update, CellConfig):
-            update = dataclasses.asdict(update)
-        new_config = dataclasses.asdict(
-            CellConfig.from_dict(deep_merge(dataclasses.asdict(self), update))
-        )
-        for key, value in new_config.items():
+            update = asdict(update)
+        new_config = CellConfig.from_dict(deep_merge(self.asdict(), update))
+        for key, value in new_config.asdict().items():
             self.__setattr__(key, value)
 
 
 CellConfigKeys = frozenset(
-    {field.name for field in dataclasses.fields(CellConfig)}
+    {field.name for field in msgspec.structs.fields(CellConfig)}
 )
 
 
@@ -156,6 +157,8 @@ class CellImpl:
     mod: ast.Module
     defs: set[Name]
     refs: set[Name]
+    # metadata about refs, currently only tracks SQL refs
+    sql_refs: dict[Name, SQLRef]
     # Variables that should only live for the duration of the cell
     temporaries: set[Name]
 
