@@ -124,25 +124,17 @@ def to_decorator(
     config: Optional[CellConfig],
     fn: Decorators = "cell",
 ) -> str:
-    if config is None:
+    if config is None or not config.is_different_from_default():
         return f"@app.{fn}"
 
-    # Removed defaults. If the cell's config is the default config,
-    # don't include it in the decorator.
-    if not config.disabled:
-        del config.disabled
-    if not config.hide_code:
-        del config.hide_code
-    if not isinstance(config.column, int):
-        del config.column
-
-    if config == CellConfig():
-        return f"@app.{fn}"
-    else:
-        return format_tuple_elements(
-            f"@app.{fn}(...)",
-            tuple(f"{key}={value}" for key, value in config.__dict__.items()),
-        )
+    # Only include non-defaults in the decorator call
+    return format_tuple_elements(
+        f"@app.{fn}(...)",
+        tuple(
+            f"{key}={value}"
+            for key, value in config.asdict_without_defaults().items()
+        ),
+    )
 
 
 def build_setup_section(setup_cell: Optional[CellImpl]) -> str:
