@@ -16,6 +16,7 @@ from marimo._runtime.packages.module_name_to_pypi_name import (
 )
 from marimo._runtime.packages.package_manager import (
     CanonicalizingPackageManager,
+    LogCallback,
     PackageDescription,
 )
 from marimo._runtime.packages.utils import split_packages
@@ -55,13 +56,19 @@ class PipPackageManager(PypiPackageManager):
     name = "pip"
     docs_url = "https://pip.pypa.io/"
 
-    async def _install(self, package: str, *, upgrade: bool) -> bool:
+    async def _install(
+        self,
+        package: str,
+        *,
+        upgrade: bool,
+        log_callback: Optional[LogCallback] = None,
+    ) -> bool:
         LOGGER.info(f"Installing {package} with pip")
         cmd = ["pip", "--python", PY_EXE, "install"]
         if upgrade:
             cmd.append("--upgrade")
         cmd.extend(split_packages(package))
-        return self.run(cmd)
+        return self.run(cmd, log_callback=log_callback)
 
     async def uninstall(self, package: str) -> bool:
         LOGGER.info(f"Uninstalling {package} with pip")
@@ -73,7 +80,8 @@ class PipPackageManager(PypiPackageManager):
                 "uninstall",
                 "-y",
                 *split_packages(package),
-            ]
+            ],
+            log_callback=None,
         )
 
     def list_packages(self) -> list[PackageDescription]:
@@ -91,7 +99,13 @@ class MicropipPackageManager(PypiPackageManager):
     def is_manager_installed(self) -> bool:
         return is_pyodide()
 
-    async def _install(self, package: str, *, upgrade: bool) -> bool:
+    async def _install(
+        self,
+        package: str,
+        *,
+        upgrade: bool,
+        log_callback: Optional[LogCallback] = None,
+    ) -> bool:
         assert is_pyodide()
         import micropip  # type: ignore
 
@@ -104,9 +118,15 @@ class MicropipPackageManager(PypiPackageManager):
                 pass
 
         try:
+            if log_callback:
+                log_callback(f"Installing {package} with micropip...\n")
             await micropip.install(split_packages(package))
+            if log_callback:
+                log_callback(f"Successfully installed {package}\n")
             return True
-        except ValueError:
+        except ValueError as e:
+            if log_callback:
+                log_callback(f"Failed to install {package}: {e}\n")
             return False
 
     async def uninstall(self, package: str) -> bool:
@@ -145,7 +165,13 @@ class UvPackageManager(PypiPackageManager):
     def is_manager_installed(self) -> bool:
         return self._uv_bin != "uv" or super().is_manager_installed()
 
-    async def _install(self, package: str, *, upgrade: bool) -> bool:
+    async def _install(
+        self,
+        package: str,
+        *,
+        upgrade: bool,
+        log_callback: Optional[LogCallback] = None,
+    ) -> bool:
         install_cmd: list[str]
         if self.is_in_uv_project:
             LOGGER.info(f"Installing in {package} with 'uv add'")
@@ -165,7 +191,9 @@ class UvPackageManager(PypiPackageManager):
 
         return self.run(
             # trade installation time for faster start time
-            install_cmd + ["--compile", *split_packages(package), "-p", PY_EXE]
+            install_cmd
+            + ["--compile", *split_packages(package), "-p", PY_EXE],
+            log_callback=log_callback,
         )
 
     def update_notebook_script_metadata(
@@ -177,7 +205,7 @@ class UvPackageManager(PypiPackageManager):
         import_namespaces_to_add: Optional[list[str]] = None,
         import_namespaces_to_remove: Optional[list[str]] = None,
         upgrade: bool,
-    ) -> None:
+    ) -> bool:
         """Update the notebook's script metadata with the packages to add/remove.
 
         Args:
@@ -201,7 +229,7 @@ class UvPackageManager(PypiPackageManager):
         ]
 
         if not packages_to_add and not packages_to_remove:
-            return
+            return True
 
         LOGGER.info(f"Updating script metadata for {filepath}")
 
@@ -243,7 +271,7 @@ class UvPackageManager(PypiPackageManager):
         packages_to_add: list[str],
         packages_to_remove: list[str],
         upgrade: bool,
-    ) -> None:
+    ) -> bool:
         from marimo._convert.markdown.markdown import extract_frontmatter
         from marimo._utils import yaml
         from marimo._utils.inline_script_metadata import (
@@ -269,7 +297,7 @@ class UvPackageManager(PypiPackageManager):
             temp_file.write(header)
             temp_file.flush()
         # Have UV modify it
-        self._process_changes_for_script_metadata(
+        result = self._process_changes_for_script_metadata(
             temp_file.name,
             packages_to_add,
             packages_to_remove,
@@ -299,24 +327,29 @@ class UvPackageManager(PypiPackageManager):
         with open(filepath, "w", encoding="utf-8") as f:
             f.write("\n".join(document))
 
+        return result
+
     def _process_changes_for_script_metadata(
         self,
         filepath: str,
         packages_to_add: list[str],
         packages_to_remove: list[str],
         upgrade: bool,
-    ) -> None:
+    ) -> bool:
+        success = True
         if packages_to_add:
             cmd = [self._uv_bin, "--quiet", "add", "--script", filepath]
             if upgrade:
                 cmd.append("--upgrade")
             cmd.extend(packages_to_add)
-            self.run(cmd)
+            success &= self.run(cmd, log_callback=None)
         if packages_to_remove:
-            self.run(
+            success &= self.run(
                 [self._uv_bin, "--quiet", "remove", "--script", filepath]
-                + packages_to_remove
+                + packages_to_remove,
+                log_callback=None,
             )
+        return success
 
     def _get_version_map(self) -> dict[str, str]:
         packages = self.list_packages()
@@ -374,10 +407,31 @@ class UvPackageManager(PypiPackageManager):
             uninstall_cmd = [self._uv_bin, "pip", "uninstall"]
 
         return self.run(
-            uninstall_cmd + [*split_packages(package), "-p", PY_EXE]
+            uninstall_cmd + [*split_packages(package), "-p", PY_EXE],
+            log_callback=None,
         )
 
     def list_packages(self) -> list[PackageDescription]:
+        # First try with `uv tree`
+        tree = self.dependency_tree()
+        if tree is not None:
+            LOGGER.info("Listing packages with 'uv tree'")
+            seen: set[str] = set()
+            packages: list[PackageDescription] = []
+            stack = list(tree.dependencies)
+            while stack:
+                pkg = stack.pop()
+                if pkg.name not in seen:
+                    packages.append(
+                        PackageDescription(
+                            name=pkg.name, version=pkg.version or ""
+                        )
+                    )
+                    seen.add(pkg.name)
+                    # Add dependencies to stack for recursion
+                    stack.extend(pkg.dependencies)
+            return sorted(packages, key=lambda pkg: pkg.name)
+
         LOGGER.info("Listing packages with 'uv pip list'")
         cmd = [self._uv_bin, "pip", "list", "--format=json", "-p", PY_EXE]
         return self._list_packages_from_cmd(cmd)
@@ -420,15 +474,26 @@ class RyePackageManager(PypiPackageManager):
     name = "rye"
     docs_url = "https://rye.astral.sh/"
 
-    async def _install(self, package: str, *, upgrade: bool) -> bool:
+    async def _install(
+        self,
+        package: str,
+        *,
+        upgrade: bool,
+        log_callback: Optional[LogCallback] = None,
+    ) -> bool:
         if upgrade:
             return self.run(
-                ["rye", "sync", "--update", *split_packages(package)]
+                ["rye", "sync", "--update", *split_packages(package)],
+                log_callback=log_callback,
             )
-        return self.run(["rye", "add", *split_packages(package)])
+        return self.run(
+            ["rye", "add", *split_packages(package)], log_callback=log_callback
+        )
 
     async def uninstall(self, package: str) -> bool:
-        return self.run(["rye", "remove", *split_packages(package)])
+        return self.run(
+            ["rye", "remove", *split_packages(package)], log_callback=None
+        )
 
     def list_packages(self) -> list[PackageDescription]:
         cmd = ["rye", "list", "--format=json"]
@@ -439,7 +504,13 @@ class PoetryPackageManager(PypiPackageManager):
     name = "poetry"
     docs_url = "https://python-poetry.org/docs/"
 
-    async def _install(self, package: str, *, upgrade: bool) -> bool:
+    async def _install(
+        self,
+        package: str,
+        *,
+        upgrade: bool,
+        log_callback: Optional[LogCallback] = None,
+    ) -> bool:
         if upgrade:
             return self.run(
                 [
@@ -447,16 +518,19 @@ class PoetryPackageManager(PypiPackageManager):
                     "update",
                     "--no-interaction",
                     *split_packages(package),
-                ]
+                ],
+                log_callback=log_callback,
             )
 
         return self.run(
-            ["poetry", "add", "--no-interaction", *split_packages(package)]
+            ["poetry", "add", "--no-interaction", *split_packages(package)],
+            log_callback=log_callback,
         )
 
     async def uninstall(self, package: str) -> bool:
         return self.run(
-            ["poetry", "remove", "--no-interaction", *split_packages(package)]
+            ["poetry", "remove", "--no-interaction", *split_packages(package)],
+            log_callback=None,
         )
 
     def _list_packages_from_cmd(
