@@ -19,7 +19,8 @@ from marimo._config.config import (
 )
 from marimo._server.ai.constants import DEFAULT_MAX_TOKENS, DEFAULT_MODEL
 from marimo._server.ai.ids import AiModelId
-from marimo._server.ai.tools import Tool, get_tool_manager
+from marimo._server.ai.tools.tool_manager import get_tool_manager
+from marimo._server.ai.tools.types import Tool
 from marimo._server.api.status import HTTPStatus
 
 
